"""Umbrella API for R2X model."""

# System packages
import importlib
import inspect
import sys

from loguru import logger

from infrasys.system import System
from r2x.exporter.handler import get_exporter

# from r2x.models import *
# Module level imports
from .config import Configuration, Scenario
from .exporter import exporter_list
from .parser import parser_list
from .parser.handler import BaseParser, get_parser_data
from .utils import (
    DEFAULT_PLUGIN_PATH,
)


def run_parser(config: Scenario, **kwargs):
    """Call get parser for parser selected.

    The parser gets selected based on `config.input_model`.

    Parameters
    ----------
    config
        Scenario
    upgrader
        Function to upgrade the inputs

    Other Parameters
    ----------------
    kwargs
        Additional key arguments to the parser.

    Raises
    ------
    KeyError
        If parser is not found on parser_list
    """
    # At some point we will read the ReEDS tag and upgrade accordingly.
    # reeds_meta = get_file(fname="meta.csv", config=config).iloc[:1]
    assert config.run_folder
    assert config.input_model
    if getattr(config, "upgrade", None):
        # Here is the logic to run the upgrader
        # upgrader(config.run_folder, save=True)
        pass

    # Initialize parser
    parser_class = parser_list.get(config.input_model)
    if not parser_class:
        raise KeyError(f"Parser for {config.input_model} not found")

    parser = get_parser_data(config, parser_class, **kwargs)
    system = parser.build_system()

    assert system is not None, "System failed to create"

    # match config.input_model:
    #     case "reeds-US":
    #         return system
    #     case "plexos" | "sienna":
    #         return parser
    #     case _:
    #         raise NotImplementedError(f"Parser for {config.input_model} is not defined.")
    return system, parser


def run_plugins(config: Scenario, parser: BaseParser, system: System) -> System:
    """Run selected plugins.

    Parameters
    ----------
    config
        Configuration manager
    system
        System

    Returns
    -------
        System
    """
    if not config.plugins:
        return system

    logger.info("Running the following plugins: {}", config.plugins)
    for plugin in config.plugins:
        module = importlib.import_module(f".{plugin}", DEFAULT_PLUGIN_PATH)
        if hasattr(module, "update_system"):
            plugin_required_args = inspect.getfullargspec(module.update_system).args
            plugin_config_args = {
                key: value for key, value in config.__dict__.items() if key in plugin_required_args
            }
            system = module.update_system(config=config, parser=parser, system=system, **plugin_config_args)
    return system


def run_exporter(config: Scenario, system: System) -> None:
    """Create exporter model."""
    assert config.output_model
    exporter_class = exporter_list.get(config.output_model, None)
    if not exporter_class:
        raise KeyError(f"Exporter for {config.output_model} not found")

    get_exporter(config, system, exporter_class)


def run_single_scenario(scenario: Scenario, **kwargs) -> None:
    """Run translation process."""
    logger.info("Running {}", scenario.name)

    if scenario.input_model == "infrasys":
        fname = f"{scenario.run_folder}/{scenario.name}.json"
        system = System.from_json(filename=fname, **kwargs)
    else:
        system, parser = run_parser(scenario, **kwargs)
        system = run_plugins(config=scenario, parser=parser, system=system)

    if getattr(scenario, "inspect", None):
        from IPython import embed

        embed()
        sys.exit(0)

    # Serialize/deserialize the system
    output_fpath = str(scenario.output_folder) + f"/{scenario.name}.json"
    if getattr(scenario, "save", None):
        logger.info("Serialize system to {}", output_fpath)
        system.to_json(output_fpath, overwrite=True)
        system = System.from_json(output_fpath)

    run_exporter(config=scenario, system=system)
    return


def scenario_runner(config_mgr: Configuration, **kwargs) -> None:
    """Run the different translation for a configuration.

    If the config object is Configuration class, we run all the different
    Scenarios in parallel using multiprocessing.

    Parameters
    ----------
    config
        Configuration

    Other Parameters
    ----------------
    kwargs
        arguments passed for convenience.
    """
    logger.debug("Running {} scenarios", len(config_mgr))
    for _, scenario in config_mgr.scenarios.items():
        if not isinstance(scenario.solve_year, list) or len(scenario.solve_year) == 1:
            run_single_scenario(scenario, **kwargs)
    return



# ruff: noqa
"""Set of helper functions for parsers."""

from loguru import logger
import polars as pl
import pandas as pd
from datetime import datetime


def pl_filter_year(df, year: int | None = None, year_columns=["t", "year"], **kwargs):
    if year is None and kwargs.get("solve_year"):
        year = kwargs["solve_year"]

    if year is None:
        return df
    matching_names = list(set(year_columns).intersection(df.collect_schema()))
    if not matching_names:
        return df

    if len(matching_names) > 1:
        return KeyError(f"More than one column identified as year. {matching_names=}")
    logger.trace("Filtering data for year {}", year)
    return df.filter(pl.col(matching_names[0]) == year)


def field_filter(property_fields, eligible_fields):
    valid = {k: v for k, v in property_fields.items() if k in eligible_fields if v is not None}
    extra = {k: v for k, v in property_fields.items() if k not in eligible_fields if v is not None}

    return valid, extra


def prepare_ext_field(valid_fields, extra_fields):
    """Cleanses the extra fields by removing any timeseries data"""
    if extra_fields:
        # Implement any filtering of ext_data here
        valid_fields["ext"] = extra_fields
    else:
        valid_fields["ext"] = {}
    return valid_fields


def filter_property_dates(system_data: pl.DataFrame, study_year: int):
    """filters query by date_from and date_to"""
    # Convert date_from and date_to to datetime
    system_data = system_data.with_columns(
        [
            pl.col("date_from").str.strptime(pl.Datetime, "%Y-%m-%dT%H:%M:%S").cast(pl.Date),
            pl.col("date_to").str.strptime(pl.Datetime, "%Y-%m-%dT%H:%M:%S").cast(pl.Date),
        ]
    )

    # Create two new columns year_from and year_to
    system_data = system_data.with_columns(
        [
            pl.col("date_from").dt.year().alias("year_from"),
            pl.col("date_to").dt.year().alias("year_to"),
        ]
    )

    # Remove Property by study year & date_from/to
    date_filter = ((pl.col("date_from").is_null()) | (pl.col("year_from") <= study_year)) & (
        (pl.col("date_to").is_null()) | (pl.col("year_to") >= study_year)
    )
    system_data = system_data.filter(date_filter)
    return system_data


def pl_lowercase(df: pl.DataFrame, **kwargs):
    logger.trace("Lowercase columns: {}", df.collect_schema().names())
    result = df.with_columns(pl.col(pl.String).str.to_lowercase()).rename(
        {column: column.lower() for column in df.collect_schema().names()}
    )
    logger.trace("New columns: {}", df.collect_schema().names())
    return result


def pl_rename(
    df: pl.DataFrame,
    column_mapping: dict[str, str] | None = None,
    default_mapping: dict[str, str] | None = None,
    **kwargs,
):
    if not column_mapping:
        logger.trace("No column mapping provided")
        return df
    logger.trace("Apply column mapping {}", column_mapping)
    return df.select(pl.col(column_mapping.keys())).select(
        pl.all().name.map(lambda col_name: column_mapping.get(col_name, col_name))
    )


def pl_left_multi_join(l_df: pl.LazyFrame, *r_dfs: pl.LazyFrame, **kwargs):
    original_keys = set(l_df.collect_schema().names())
    # output_df = pl.LazyFrame(l_df)
    output_df = l_df
    for r_df in r_dfs:
        current_keys = set(r_df.collect_schema().names())
        current_keys = original_keys.intersection(current_keys)
        output_df = output_df.join(r_df, on=list(current_keys), how="left", coalesce=True)

    output_df = output_df.collect()
    assert (
        output_df.shape[0] == l_df.collect().shape[0]
    ), f"Merge resulted in less rows. Check the shared keys. {original_keys=} vs {current_keys=}"
    return output_df


def handle_leap_year_adjustment(data_file: pl.DataFrame):
    """Duplicate feb 28th to feb 29th for leap years."""
    feb_28 = data_file.slice(1392, 24)
    before_feb_29 = data_file.slice(0, 1416)
    after_feb_29 = data_file.slice(1416, len(data_file) - 1440)
    return pl.concat([before_feb_29, feb_28, after_feb_29])


def fill_missing_timestamps(data_file: pl.DataFrame, date_time_column: list[str]):
    """Add missing timestamps to data and forward fill nulls"""
    data_file = data_file.with_columns(
        (
            pl.col("year").cast(pl.Int32).cast(pl.Utf8)
            + "-"
            + pl.col("month").cast(pl.Int32).cast(pl.Utf8).str.zfill(2)
            + "-"
            + pl.col("day").cast(pl.Int32).cast(pl.Utf8).str.zfill(2)
            + " "
            + pl.col("hour").cast(pl.Int32).cast(pl.Utf8).str.zfill(2)
            + ":00:00"
        )
        .str.strptime(pl.Datetime, "%Y-%m-%d %H:%M:%S")
        .alias("timestamp")
    ).with_columns(pl.col("timestamp").dt.cast_time_unit("ns"))

    data_file = data_file.with_columns(
        pl.col("year").cast(pl.Int32),
        pl.col("month").cast(pl.Int8),
        pl.col("day").cast(pl.Int8),
        pl.col("hour").cast(pl.Int8),
    )

    complete_timestamps = pl.from_pandas(pd.DataFrame({"timestamp": date_time_column}))
    missing_timestamps = complete_timestamps.join(data_file, on="timestamp", how="anti")

    missing_timestamps = missing_timestamps.with_columns(
        pl.col("timestamp").dt.year().alias("year"),
        pl.col("timestamp").dt.month().alias("month"),
        pl.col("timestamp").dt.day().alias("day"),
        pl.col("timestamp").dt.hour().alias("hour"),
        pl.lit(None).alias("value"),
    ).select(["year", "month", "day", "hour", "value", "timestamp"])

    complete_df = pl.concat([data_file, missing_timestamps]).sort("timestamp").fill_null(strategy="forward")
    complete_df.drop_in_place("timestamp")
    return complete_df


def resample_data_to_hourly(data_file: pl.DataFrame):
    """Resample data to hourly frequency from 30 minute data."""
    data_file = data_file.with_columns((pl.col("hour") % 48).alias("hour"))
    data_file = (
        data_file.with_columns(
            (
                pl.datetime(
                    data_file["year"],
                    data_file["month"],
                    data_file["day"],
                    hour=data_file["hour"] // 2,
                    minute=(data_file["hour"] % 2) * 30,
                )
            ).alias("timestamp")
        )
        .sort("timestamp")
        .filter(pl.col("timestamp").is_not_null())
    )

    return (
        data_file.group_by_dynamic("timestamp", every="1h")
        .agg([pl.col("value").mean().alias("value")])
        .with_columns(
            pl.col("timestamp").dt.year().alias("year"),
            pl.col("timestamp").dt.month().alias("month"),
            pl.col("timestamp").dt.day().alias("day"),
            pl.col("timestamp").dt.hour().alias("hour"),
            pl.col("value").alias("value"),
        )
        .select(["year", "month", "day", "hour", "value"])
    )



from r2x.enums import PrimeMoversType
from r2x.models import Generator
from r2x.plugins.break_gens import break_generators
from .models import ieee5bus


def test_break_generators():
    system = ieee5bus()
    capacity_threshold = 10
    reference_generators = {
        "storage": {"avg_capacity_MW": 100},
    }
    system = break_generators(system, reference_generators, capacity_threshold, break_category="category")
    updated_generators = list(system.get_components(Generator))
    assert len(updated_generators) == 9  # 8 original generator + 1 new one

    new_generators = system.get_components(
        Generator, filter_func=lambda x: x.prime_mover_type == PrimeMoversType.BA
    )
    for generator in new_generators:
        assert generator.active_power.magnitude == 100
        assert generator.ext["original_capacity"].magnitude == 200
        assert generator.ext["broken"]


def test_break_generators_break_category():
    system = ieee5bus()
    capacity_threshold = 10
    reference_generators = {
        "Battery1": {"avg_capacity_MW": 100},
    }
    non_break_techs = ["storage", "Solitude", "Park City"]
    system = break_generators(
        system, reference_generators, capacity_threshold, non_break_techs, break_category="name"
    )

    updated_generators = list(system.get_components(Generator))
    assert len(updated_generators) == 9  # 8 original generator + 1 new ones


def test_break_generators_multi_category():
    system = ieee5bus()
    capacity_threshold = 10
    reference_generators = {
        "storage": {"avg_capacity_MW": 100},
        "thermal": {"avg_capacity_MW": 100},
        "solar": {"avg_capacity_MW": 100},
    }
    system = break_generators(system, reference_generators, capacity_threshold)

    updated_generators = list(system.get_components(Generator))
    assert len(updated_generators) == 28  # 8 original generator + 18 new ones


def test_break_generators_capacity_threshold():
    system = ieee5bus()
    capacity_threshold = 50
    reference_generators = {
        "storage": {"avg_capacity_MW": 150},
    }
    system = break_generators(system, reference_generators, capacity_threshold)

    updated_generators = list(system.get_components(Generator))
    assert len(updated_generators) == 8  # 8 original generator - 1 dropped



"""Custom exceptions for R2X."""
# ruff: noqa: D101


class DuplicateUUIDError(Exception):
    pass


class DuplicateNameError(Exception):
    pass


class NotFoundError(Exception):
    pass


class MultlipleElementsError(Exception):
    pass


class ModelError(Exception):
    pass


class MultipleFilesError(Exception):
    pass


class ParserError(Exception):
    pass



"""Configuration handler.

This script provides the base class to save the metadata for a given ReEDS scenario.
It can either read the information directly or throught a cases file.
"""

# System packages
import csv
import inspect
import os
import pathlib
from collections import ChainMap
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

# Third-party packages
from loguru import logger
import rich
from rich.table import Table

# Local imports
from .utils import get_defaults, get_project_root, match_input_model, update_dict, validate_string


@dataclass
class Scenario:
    """Scenario class.

    The purpose of the scenario class is to provide a way to track
    the inputs and defaults for a given translation. It serves as an access
    point to get the defaults, configuration and features flags for different
    steps (e.g., parser, exporter or validation) since most of the codebase use
    object composition.

    Each scenario can have its own set of flags that will run different parts
    of the parsers.

    A very simple use case for this class::

        from r2x.config import Scenario

        scenario = Scenario(name="test")
        print(scenario)

    Parameters
    ----------
    name
        Name for the scenario
    run_folder
        Path for the scenario inputs
    output_folder
        Path for output exports
    input_model
        Model to translate from
    output_model
        model to translate to
    fmap
        Dictionary with files configuration
    defaults
        Default configuration from the scenario
    feature_flags
        Dictionary with experimental features
    plugins
        List of plugins enabled

    See Also
    --------
    Configuration
    """

    name: str | None = None
    run_folder: Path | str | None = None
    output_folder: Path | str = Path(".")
    solve_year: list[int] | int | None = None
    weather_year: int | None = None
    input_model: str | None = None
    output_model: str | None = None
    cases_flags: dict[str, Any] = field(default_factory=dict)
    fmap: dict[str, Any] = field(default_factory=dict)
    defaults: dict[str, Any] = field(default_factory=dict)
    feature_flags: dict[str, Any] = field(default_factory=dict)
    plugins: list[Any] | None = None
    _mkdir: bool = True

    def __str__(self) -> str:
        _str = (
            f"Scenario(name={self.name}, input_model={self.input_model}, output_model={self.output_model},"
            + f"run_folder={self.run_folder})"
        )
        return _str

    def __repr__(self) -> str:
        _str = (
            f"Scenario(name={self.name}, input_model={self.input_model}, output_model={self.output_model},"
            + f"run_folder={self.run_folder})"
        )
        return _str

    def info(self) -> None:
        """Return table summary of the configuration."""
        config_table = Table(
            title="R2X configuration",
            show_header=True,
            title_justify="left",
            title_style="bold",
        )
        config_table.add_column("Property", style="green", justify="left", min_width=20)
        config_table.add_column("Value", justify="left", min_width=20)

        for _field in self.__dict__:
            value = getattr(self, _field)
            if _field == "fmap":
                value = f"Length: {len(value)}"
            elif _field == "defaults":
                value = f"Length: {len(value)}"
            if value:
                config_table.add_row(_field, str(value))

        return rich.print(config_table)

    def __post_init__(self):
        self._normalize_path()
        self._set_scenario_name()
        self._load_fmap_config()
        self._load_defaults()

    def _normalize_path(self) -> None:
        if isinstance(self.output_folder, str):
            self.output_folder = pathlib.Path(self.output_folder)

        # Create output folder if it does not exists.
        if self._mkdir:
            logger.trace(f"Creating folder {self.output_folder=}.")
            self.output_folder.mkdir(exist_ok=True)
        return

    def _set_scenario_name(self) -> None:
        if not self.name and self.run_folder:
            self.name = os.path.basename(self.run_folder)
        return

    def _load_fmap_config(self) -> None:
        if self.input_model and not self.fmap:
            self.fmap = match_input_model(self.input_model)
            logger.debug(f"Getting fmap configuration for {self.input_model}")

        return

    def _load_defaults(self) -> None:
        logger.info("Getting defaults for scenario {}", self.name)
        self.defaults = get_defaults(self.input_model, self.output_model)
        if not self.plugins:
            logger.debug("Using default plugins.")
            setattr(self, "plugins", self.defaults.get("default_plugins"))
        else:
            logger.debug("Using the following plugins: {}", self.plugins)

    @classmethod
    def from_kwargs(cls, **kwargs) -> "Scenario":
        """Create Scenario instance from key arguments."""
        cls_fields = {field for field in inspect.signature(cls).parameters}

        native_args, new_args = {}, {}
        for name, val in kwargs.items():
            if name in cls_fields:
                native_args[name] = val
            else:
                new_args[name] = val

        ret = cls(**native_args)

        for new_name, new_val in new_args.items():
            setattr(ret, new_name, new_val)
        return ret


@dataclass
class Configuration:
    """R2X configuration manager that wraps multiple Scenario instances.

    This class parses either the cases_*.csv file or reads the inputs from the CLI.

    Attributes
    ----------
    scenario_list
        List of Scenario instances
    scenario_names
        Names of the scenarios in `scenario_list`

    See Also
    --------
    Scenario
    """

    scenarios: dict = field(default_factory=dict)
    scenario_names: list = field(default_factory=list)

    def __len__(self) -> int:
        """Get the number of scenarios in the class."""
        return len(self.scenarios)

    def __getitem__(self, scenario_name: str) -> Scenario:
        """Return a single scenario from the configuration.

        Parameters
        ----------
        name
            Name of the scenario

        Returns
        -------
            Scenario

        Raises
        ------
        KeyError
            If scenario does not exists.
        """
        for scenario in self.scenarios:
            if scenario == scenario_name:
                return self.scenarios[scenario]
        raise KeyError(f"No scenario named '{scenario_name}'")

    def __iter__(self):
        return iter(self.scenarios)

    def get(self, name: str) -> "Configuration":
        """Return a scenario from the configuration manager.

        Parameters
        ----------
        name
            Name of the scenario to get

        Returns
        -------
            Configuration

        Raises
        ------
        KeyError
            If name is not present on scenario_names
        """
        if name not in self.scenario_names:
            msg = f"{name=} not found in scenarios. List of available scenarios {self.scenario_names}"
            raise KeyError(msg)

        return self.scenarios[name]

    @staticmethod
    def override(params: dict, cli_args: dict, user_dict: dict | None = None) -> Scenario:
        """Override scenario either by user provided dict or cli.

        If the user want to override defaults it can be done by passing the
        `user_dict`. Currently, we only use the `user_dict` to override the defaults and the fmap.

        Parameters
        ----------
        params
            Parameters to create the scenario
        user_dict
            Dictionary that will override the defaults
        cli_args
            Arguments passed from the CLI
        """
        if user_dict is None:
            user_dict = {}
        if not any(key in params for key in cli_args) and not user_dict:
            return Scenario.from_kwargs(**params)

        # Order is CLI -> user_dict -> Default params
        logger.debug("Overriding scenario with {}", user_dict)
        override_dict = user_dict.copy()
        fmap_dict = override_dict.pop("fmap", {})
        mapping = ChainMap(cli_args, override_dict, params)
        config = Scenario.from_kwargs(**mapping)

        config.fmap = update_dict(base_dict=config.fmap, override_dict=ChainMap(cli_args, fmap_dict))
        config.defaults = update_dict(
            base_dict=config.defaults,
            override_dict=ChainMap(cli_args, override_dict),
        )
        return config

    @classmethod
    def from_cli(cls, cli_args: dict, user_dict: dict | None = None, **kwargs):
        """Create scenario from the CLI arguments.

        It saves the created scenario in the scenario_list` and scenario_names

        Parameters
        ----------
        kwargs
            Arguments for constructing the scenario.

        See Also
        --------
            Scenario.from_kwargs
        """
        instance = cls()
        scenario = Scenario.from_kwargs(**cli_args)
        if user_dict:
            scenario = instance.override(scenario.__dict__, cli_args=cli_args, user_dict=user_dict)
        instance.scenario_names.append(scenario.name)
        instance.scenarios[scenario.name] = scenario
        return instance

    @classmethod
    def from_cases(cls, cases_fpath: str, cli_args: dict, user_dict: dict | None = None):
        """Parse the legacy cases file into a list of Scenarios.

        Parameters
        ----------
        cases_fpath
            fpath of cases file

        Raises
        ------
        FileNotFoundError
            If csv file does not exists
        """
        instance = cls()
        logger.info("Parsing configuration from: {}", cases_fpath)

        # Check if only the file name is passed. If so, try to find the file in
        # the current folder
        if not cases_fpath.endswith(".csv"):
            project_path = str(get_project_root())
            cases_fpath = os.path.join(project_path, f"cases_{cases_fpath}.csv")

        with open(cases_fpath) as csv_file:
            csv_reader = csv.reader(csv_file)
            header = next(csv_reader)
            rows = list(csv_reader)
            for col_num in range(3, len(header)):
                scenario_flags = {}
                feature_flag = False
                feature_flags_dict = {}
                for row in rows[1:]:
                    flag_name = row[0]

                    # Skip uppercase flags since they are separators
                    if flag_name.isupper() and "EXPERIMENTAL" in row:
                        feature_flag = True
                        continue

                    default_value = row[1]
                    flag_value = row[col_num]

                    # Use validated default value if None
                    value = default_value if not flag_value else flag_value
                    validated_value = validate_string(value)

                    if not feature_flag:
                        scenario_flags[flag_name] = validated_value
                    else:
                        # Only save feature flag if passed. Not sure if this will stay.
                        if validated_value is not None:
                            feature_flags_dict.update({flag_name: validated_value})

                scenario_fpath = scenario_flags.get("scenario")
                run_folder = scenario_flags.get("run_folder")
                run_folder = os.path.join(str(run_folder), str(scenario_fpath))
                scenario_flags["run_folder"] = run_folder
                scenario_flags["name"] = header[col_num]

                scenario_flags = {key: value for key, value in scenario_flags.items() if value}
                scenario_flags["feature_flags"] = feature_flags_dict
                instance.scenarios[header[col_num]] = instance.override(
                    scenario_flags,
                    cli_args=cli_args,
                    user_dict=user_dict,
                )
                instance.scenario_names.append(header[col_num])
        return instance



"""Helper function to enable plugins/extension on R2X."""

from importlib.util import find_spec, module_from_spec

from loguru import logger

# Module level imports
from ..utils import DEFAULT_PLUGIN_PATH


def valid_plugin_list(plugin_list: list[str], folder: str = DEFAULT_PLUGIN_PATH, logger=logger) -> list[str]:
    """Return valid plugins for the R2X process.

    Args:
        plugin_list: list of plugins to include.
        folder: Folder that contains the plugin.
        logger: Logger object.

    Returns
    -------
        List[str] of valid plugins.

    Note(s):
        - Currently only works if the folder is located inside of the R2X
          project folder.
    """
    # Initialize list of valid plugins.
    valid_plugins = []
    for plugin in plugin_list:
        if validate_plugin(plugin, folder, logger):
            valid_plugins.append(plugin)
    if len(valid_plugins) != plugin_list:
        logger.debug(f"Updated list {valid_plugins=}")
    return valid_plugins


def validate_plugin(
    plugin: str,
    folder: str,
    logger=logger,
    valid_fn_names: list[str] = ["cli_arguments", "translator", "exporter"],
) -> bool:
    """Check if the plugin is callable and has the default function names.

    Args:
        plugin: Name of the plugin.
        folder: Folder that contains the plugin.
        logger: Logger object.
        valid_fn_names: Function names to look for function

    Returns
    -------
        True or False of the plugin is found and has any callable function names.
    """
    # Return if the plugin is not found.
    if not (spec := find_spec(f"{folder}.{plugin}")):
        logger.warning(f"Plugin {plugin} specified does not exists. Skipping it.")
        return False

    # Load plugin.
    module = module_from_spec(spec)
    if spec.loader is None:
        return False
    spec.loader.exec_module(module)

    valid_functions = [hasattr(module, fn) for fn in valid_fn_names]
    if any(valid_functions):
        return True
    else:
        logger.warning(f"{plugin} does not have any valid function names. Skipping it")
        return False



"""Models related to services."""

from typing import Annotated

from pydantic import Field, NonNegativeFloat, PositiveFloat

from r2x.enums import ReserveDirection, ReserveType
from r2x.models.core import MinMax, Service
from r2x.models.topology import LoadZone
from r2x.units import EmissionRate


class Reserve(Service):
    """Class representing a reserve contribution."""

    time_frame: Annotated[
        PositiveFloat,
        Field(description="Timeframe in which the reserve is required in seconds"),
    ] = 1e30
    region: (
        Annotated[
            LoadZone,
            Field(description="LoadZone where reserve requirement is required."),
        ]
        | None
    ) = None
    vors: Annotated[
        float,
        Field(description="Value of reserve shortage in $/MW. Any positive value as as soft constraint."),
    ] = -1
    duration: (
        Annotated[
            PositiveFloat,
            Field(description="Time over which the required response must be maintained in seconds."),
        ]
        | None
    ) = None
    reserve_type: ReserveType
    load_risk: (
        Annotated[
            NonNegativeFloat,
            Field(
                description="Proportion of Load that contributes to the requirement.",
            ),
        ]
        | None
    ) = None
    # ramp_rate: float | None = None  # NOTE: Maybe we do not need this.
    max_requirement: float = 0  # Should we specify which variable is the time series for?
    direction: ReserveDirection

    @classmethod
    def example(cls) -> "Reserve":
        return Reserve(
            name="ExampleReserve",
            region=LoadZone.example(),
            direction=ReserveDirection.UP,
            reserve_type=ReserveType.REGULATION,
        )


class Emission(Service):
    """Class representing an emission object that is attached to generators."""

    rate: Annotated[EmissionRate, Field(description="Amount of emission produced in kg/MWh.")]
    emission_type: Annotated[str, Field(description="Type of emission. E.g., CO2, NOx.")]
    generator_name: Annotated[str, Field(description="Generator emitting.")]

    @classmethod
    def example(cls) -> "Emission":
        return Emission(
            name="ExampleEmission",
            generator_name="gen1",
            rate=EmissionRate(105, "kg/MWh"),
            emission_type="CO2",
        )


class TransmissionInterface(Service):
    """A collection of branches that make up an interface or corridor for the transfer of power
    such as between different :class:Area or :class:LoadZone.

    The interface can be used to constrain the power flow across it
    """

    active_power_flow_limits: Annotated[
        MinMax, Field(description="Minimum and maximum active power flow limits on the interface (MW)")
    ]
    direction_mapping: Annotated[
        dict[str, int],
        Field(
            description=(
                "Dictionary of the line names in the interface and their direction of flow (1 or -1) "
                "relative to the flow of the interface"
            )
        ),
    ]

    @classmethod
    def example(cls) -> "TransmissionInterface":
        return TransmissionInterface(
            name="ExampleTransmissionInterface",
            active_power_flow_limits=MinMax(-100, 100),
            direction_mapping={"line-01": 1, "line-02": -2},
        )



"""Plugin to dissaggregate and aggregate generators.

This plugin breaks apart generators that are to big in conmparisson with the
WECC database. If the generator is to small after the breakup than the capacity
threshold variable, we drop the genrator entirely.
"""

# System packages
import re
from argparse import ArgumentParser
from infrasys.base_quantity import BaseQuantity
import numpy as np
import pandas as pd
from r2x.api import System

from r2x.models import Emission, Generator
from r2x.config import Scenario
from r2x.parser.handler import BaseParser
from r2x.units import ureg, ActivePower
from r2x.utils import read_json

# Local imports

from loguru import logger


# Constants
CAPACITY_THRESHOLD = 5  # MW
PROPERTIES_TO_BREAK = [
    "ramp_up",
    "ramp_down",
    "min_rated_capacity",
    "startup_cost",
    "pump_load",
    "storage_capacity",
]


def cli_arguments(parser: ArgumentParser):
    """CLI arguments for the plugin."""
    parser.add_argument(
        "--capacity-threshold",
        type=int,
        help="Capacity threshold for aggregating generators.",
        default=CAPACITY_THRESHOLD,
    )


def update_system(
    config: Scenario,
    parser: BaseParser,
    system: System,
    pcm_defaults_fpath: str | None = None,
    capacity_threshold: int = CAPACITY_THRESHOLD,
) -> System:
    """Break apart large generators based on average capacity.

    This function updates and overrides the data objects from the translator to
    be passed to the exporter.

    Args:
        translator: r2x.Translator object,
        *args: additional arguments that can be passed,
        **kwargs: additiona arguments.
    """
    logger.info("Dividing generators into average size generators")
    if pcm_defaults_fpath is None:
        logger.debug("Using {}", config.defaults["pcm_defaults_fpath"])
        pcm_defaults = read_json(config.defaults["pcm_defaults_fpath"])
    else:
        logger.debug("Using custom defaults from {}", pcm_defaults_fpath)
        pcm_defaults: dict = read_json(pcm_defaults_fpath)

    reference_generators = (
        pd.DataFrame.from_dict(pcm_defaults)
        .transpose()
        .reset_index()
        .rename(
            columns={
                "index": "tech",
            }
        )
        .set_index("tech")
        .replace({np.nan: None})
        .to_dict(orient="index")
    )
    non_break_techs = config.defaults.get("non_break_techs", [])

    return break_generators(system, reference_generators, capacity_threshold, non_break_techs)


def break_generators(  # noqa: C901
    system: System,
    reference_generators: dict[str, dict],
    capacity_threshold: int = CAPACITY_THRESHOLD,
    non_break_techs: list[str] | None = None,
    break_category: str = "category",
) -> System:
    """Break component generator into smaller units."""
    if non_break_techs:
        regex_pattern = f"^(?!{'|'.join(non_break_techs)})."  # Oh yes.
    else:
        regex_pattern = ".*"

    capacity_dropped = 0  # count capacity dropped
    for component in system.get_components(Generator, filter_func=lambda x: re.search(regex_pattern, x.name)):
        if not (tech := getattr(component, break_category, None)):
            logger.trace("Skipping component {} with missing category", component.label)
            continue
        logger.trace("Breaking {}", component.label)

        if not (reference_tech := reference_generators.get(tech)):
            logger.trace("{} not found in reference_generators", tech)
            continue
        if not (avg_capacity := reference_tech.get("avg_capacity_MW", None)):
            continue
        logger.trace("Average_capacity: {}", avg_capacity)
        reference_base_power = (
            component.active_power.magnitude
            if isinstance(component.active_power, BaseQuantity)
            else component.active_power
        )
        no_splits = int(reference_base_power // avg_capacity)
        remainder = reference_base_power % avg_capacity
        if no_splits > 1:
            split_no = 1
            logger.trace(
                "Breaking generator {} with active_power {} into {} generators of {} capacity",
                component.name,
                reference_base_power,
                no_splits,
                avg_capacity,
            )
            for _ in range(no_splits):
                component_name = component.name + f"_{split_no:02}"
                new_component = system.copy_component(component, name=component_name, attach=True)
                new_base_power = (
                    ActivePower(avg_capacity, component.active_power.units)
                    if isinstance(component.active_power, BaseQuantity)
                    else avg_capacity * ureg.MW
                )
                new_component.active_power = new_base_power
                proportion = (
                    avg_capacity / reference_base_power
                )  # Required to recalculate properties that depend on active_power
                for property in PROPERTIES_TO_BREAK:
                    if attr := getattr(new_component, property, None):
                        new_component.ext[f"{property}_original"] = attr
                        setattr(new_component, property, attr * proportion)
                new_component.ext["original_capacity"] = component.active_power
                new_component.ext["original_name"] = component.name
                new_component.ext["broken"] = True

                # NOTE: This will be migrated once we implement the SQLite for the components.
                # Add emission objects
                for emission in system.get_components(
                    Emission, filter_func=lambda x: x.generator_name == component.name
                ):
                    new_emission = system.copy_component(
                        emission, name=f"{component_name}_{emission.emission_type}", attach=True
                    )
                    new_emission.generator_name = component_name
                    system.remove_component(emission)

                if system.has_time_series(component):
                    logger.trace(
                        "Component {} has time series attached to it. Copying first one", component.label
                    )
                    ts = system.get_time_series(component)
                    system.add_time_series(ts, new_component)
                split_no += 1
            if remainder > capacity_threshold:
                component_name = component.name + f"_{split_no:02}"
                new_component = system.copy_component(component, name=component_name, attach=True)
                new_component.active_power = remainder * ureg.MW
                proportion = (
                    remainder / reference_base_power
                )  # Required to recalculate properties that depend on active_power
                for property in PROPERTIES_TO_BREAK:
                    if attr := getattr(new_component, property, None):
                        new_component.ext[f"{property}_original"] = attr
                        setattr(new_component, property, attr * proportion)
                new_component.ext["original_capacity"] = component.active_power
                new_component.ext["original_name"] = component.name
                new_component.ext["broken"] = True
                # NOTE: This will be migrated once we implement the SQLite for the components.
                # Add emission objects
                for emission in system.get_components(
                    Emission, filter_func=lambda x: x.generator_name == component.name
                ):
                    new_emission = system.copy_component(
                        emission, name=f"{component_name}_{emission.emission_type}", attach=True
                    )
                    new_emission.generator_name = component_name
                    system.remove_component(emission)
                if system.has_time_series(component):
                    logger.trace(
                        "Component {} has time series attached to it. Copying first one", component.label
                    )
                    ts = system.get_time_series(component)
                    system.add_time_series(ts, new_component)
            else:
                capacity_dropped = capacity_dropped + remainder
                logger.debug("Dropped {} capacity for {}", remainder, component.name)

            # Finally remove the component
            system.remove_component(component)
    logger.info("Total capacity dropped {} MW", capacity_dropped)
    return system



import pytest
from r2x.config import Scenario
from r2x.exporter.sienna import SiennaExporter
from .models import ieee5bus


@pytest.fixture
def infrasys_test_system():
    return ieee5bus()


@pytest.fixture
def scenario_instance(data_folder, tmp_folder):
    return Scenario(
        name="Test Scenario",
        run_folder=data_folder,
        output_folder=tmp_folder,
        input_model="infrasys",
        output_model="sienna",
        weather_year=2010,
    )


@pytest.fixture
def sienna_exporter(scenario_instance, infrasys_test_system, tmp_folder):
    return SiennaExporter(config=scenario_instance, system=infrasys_test_system, output_folder=tmp_folder)


def test_sienna_exporter_instance(sienna_exporter):
    assert isinstance(sienna_exporter, SiennaExporter)


def test_sienna_exporter_run(sienna_exporter, tmp_folder):
    exporter = sienna_exporter.run()

    output_files = [
        "gen.csv",
        "bus.csv",
        "timeseries_pointers.json",
        "storage.csv",
        # "reserves.csv",  # Reserve could be optional
        "dc_branch.csv",
        "branch.csv",
    ]

    for file in output_files:
        assert (tmp_folder / file).exists(), f"File {file} was not created properly."

    # Check that time series was created correctly
    ts_directory = tmp_folder / exporter.ts_directory
    assert any(ts_directory.iterdir())



"""Models for generator devices."""

from typing import Annotated

from pydantic import Field, NonNegativeFloat

from r2x.models.core import Device
from r2x.models.costs import HydroGenerationCost, RenewableGenerationCost, ThermalGenerationCost, StorageCost
from r2x.models.topology import ACBus
from r2x.models.load import PowerLoad
from r2x.enums import PrimeMoversType
from r2x.units import (
    ActivePower,
    FuelPrice,
    Percentage,
    PowerRate,
    ApparentPower,
    ureg,
    Time,
    Energy,
)


class Generator(Device):
    """Abstract generator class."""

    bus: Annotated[ACBus, Field(description="Bus where the generator is connected.")] | None = None
    rating: Annotated[
        ApparentPower | None,
        Field(ge=0, description="Maximum output power rating of the unit (MVA)."),
    ] = None
    active_power: Annotated[
        ActivePower,
        Field(
            description=(
                "Initial active power set point of the unit in MW. For power flow, this is the steady "
                "state operating point of the system."
            ),
        ),
    ]
    operation_cost: (
        ThermalGenerationCost | RenewableGenerationCost | HydroGenerationCost | StorageCost | None
    ) = None
    base_power: Annotated[
        ApparentPower | None,
        Field(
            gt=0,
            description="Base power of the unit (MVA) for per unitization.",
        ),
    ] = None
    must_run: Annotated[int | None, Field(description="If we need to force the dispatch of the device.")] = (
        None
    )
    vom_price: Annotated[FuelPrice, Field(description="Variable operational price $/MWh.")] | None = None
    prime_mover_type: (
        Annotated[PrimeMoversType, Field(description="Prime mover technology according to EIA 923.")] | None
    ) = None
    unit_type: Annotated[
        PrimeMoversType | None, Field(description="Prime mover technology according to EIA 923.")
    ] = None
    min_rated_capacity: Annotated[ActivePower, Field(description="Minimum rated power generation.")] = (
        0 * ureg.MW
    )
    ramp_up: (
        Annotated[
            PowerRate,
            Field(description="Ramping rate on the positve direction."),
        ]
        | None
    ) = None
    ramp_down: (
        Annotated[
            PowerRate,
            Field(description="Ramping rate on the negative direction."),
        ]
        | None
    ) = None
    min_up_time: (
        Annotated[
            Time,
            Field(ge=0, description="Minimum up time in hours for UC decision."),
        ]
        | None
    ) = None
    min_down_time: (
        Annotated[
            Time,
            Field(ge=0, description="Minimum down time in hours for UC decision."),
        ]
        | None
    ) = None
    mean_time_to_repair: (
        Annotated[
            Time,
            Field(gt=0, description="Total hours to repair after outage occur."),
        ]
        | None
    ) = None
    forced_outage_rate: (
        Annotated[
            Percentage,
            Field(description="Expected level of unplanned outages in percent."),
        ]
        | None
    ) = None
    planned_outage_rate: (
        Annotated[
            Percentage,
            Field(description="Expected level of planned outages in percent."),
        ]
        | None
    ) = None
    startup_cost: Annotated[NonNegativeFloat, Field(description="Cost in $ of starting a unit.")] | None = (
        None
    )
    shutdown_cost: (
        Annotated[NonNegativeFloat, Field(description="Cost in $ of shuting down a unit.")] | None
    ) = None


class RenewableGen(Generator):
    """Abstract class for renewable generators."""


class RenewableDispatch(RenewableGen):
    """Curtailable renewable generator.

    This type of generator have a hourly capacity factor profile.
    """


class RenewableNonDispatch(RenewableGen):
    """Non-curtailable renewable generator.

    Renewable technologies w/o operational cost.
    """


class HydroGen(Generator):
    """Hydroelectric generator."""


class HydroDispatch(HydroGen):
    """Class representing flexible hydro generators."""

    ramp_up: (
        Annotated[
            PowerRate,
            Field(ge=0, description="Ramping rate on the positve direction."),
        ]
        | None
    ) = None
    ramp_down: (
        Annotated[
            PowerRate,
            Field(ge=0, description="Ramping rate on the negative direction."),
        ]
        | None
    ) = None


class HydroFix(HydroGen):
    """Class representing unflexible hydro."""


class HydroEnergyReservoir(HydroGen):
    """Class representing hydro system with reservoirs."""

    initial_energy: (
        Annotated[NonNegativeFloat, Field(description="Initial water volume or percentage.")] | None
    ) = 0
    storage_capacity: (
        Annotated[
            Energy,
            Field(description="Total water volume or percentage."),
        ]
        | None
    ) = None
    min_storage_capacity: (
        Annotated[
            Energy,
            Field(description="Minimum water volume or percentage."),
        ]
        | None
    ) = None
    storage_target: (
        Annotated[
            Energy,
            Field(description="Maximum energy limit."),
        ]
        | None
    ) = None


class HydroPumpedStorage(HydroGen):
    """Class representing pumped hydro generators."""

    storage_duration: (
        Annotated[
            Time,
            Field(description="Storage duration in hours."),
        ]
        | None
    ) = None
    initial_volume: (
        Annotated[Energy, Field(gt=0, description="Initial water volume or percentage.")] | None
    ) = None
    storage_capacity: Annotated[
        Energy,
        Field(gt=0, description="Total water volume or percentage."),
    ]
    min_storage_capacity: (
        Annotated[
            Energy,
            Field(description="Minimum water volume or percentage."),
        ]
        | None
    ) = None
    pump_efficiency: Annotated[Percentage, Field(ge=0, le=1, description="Pumping efficiency.")] | None = None
    pump_load: (
        Annotated[
            ActivePower,
            Field(description="Load related to the usage of the pump."),
        ]
        | None
    ) = None

    @classmethod
    def example(cls) -> "HydroPumpedStorage":
        return HydroPumpedStorage(
            name="HydroPumpedStorage",
            active_power=ActivePower(100, "MW"),
            pump_load=ActivePower(100, "MW"),
            bus=ACBus.example(),
            prime_mover_type=PrimeMoversType.PS,
            storage_duration=Time(10, "h"),
            storage_capacity=Energy(1000, "MWh"),
            min_storage_capacity=Energy(10, "MWh"),
            pump_efficiency=Percentage(85, "%"),
            initial_volume=Energy(500, "MWh"),
            ext={"description": "Pumped hydro unit with 10 hour of duration"},
        )


class ThermalGen(Generator):
    """Class representing fuel based thermal generator."""

    fuel: Annotated[str, Field(description="Fuel category")] | None = None


class ThermalStandard(ThermalGen):
    """Standard representation of thermal device."""

    @classmethod
    def example(cls) -> "ThermalStandard":
        return ThermalStandard(
            name="ThermalStandard",
            bus=ACBus.example(),
            fuel="gas",
            active_power=100.0 * ureg.MW,
            ext={"Additional data": "Additional value"},
        )


class ThermalMultiStart(ThermalGen):
    """We will fill this class once we have the need for it."""


class Storage(Generator):
    """Default Storage class."""

    storage_duration: (
        Annotated[
            Time,
            Field(description="Storage duration in hours."),
        ]
        | None
    ) = None
    storage_capacity: Annotated[
        Energy,
        Field(description="Maximum allowed volume or state of charge."),
    ]
    initial_energy: Annotated[Percentage, Field(description="Initial state of charge.")] | None = None
    min_storage_capacity: Annotated[Percentage, Field(description="Minimum state of charge")] = Percentage(
        0, "%"
    )
    max_storage_capacity: Annotated[Percentage, Field(description="Minimum state of charge")] = Percentage(
        100, "%"
    )


class GenericBattery(Storage):
    """Battery energy storage model."""

    charge_efficiency: Annotated[Percentage, Field(ge=0, description="Charge efficiency.")] | None = None
    discharge_efficiency: Annotated[Percentage, Field(ge=0, description="Discharge efficiency.")] | None = (
        None
    )


class HybridSystem(Device):
    """Representation of hybrid system with renewable generation, load, thermal generation and storage.

    This class is just a link between two components.
    For the implementation see:
    - https://github.com/NREL-Sienna/PowerSystems.jl/blob/main/src/models/HybridSystem.jl
    """

    storage_unit: Storage | None = None
    renewable_unit: RenewableGen | None = None
    thermal_unit: ThermalGen | None = None
    electric_load: PowerLoad | None = None



"""Testing for the configuration and Scenario class."""

import pytest
from r2x.config import Scenario, Configuration
from r2x.utils import read_fmap


@pytest.fixture
def scenario_instance(data_folder, tmp_folder):
    return Scenario(
        name="Test Scenario",
        run_folder=data_folder,
        output_folder=tmp_folder,
        input_model="plexos",
        output_model="sienna",
    )


def test_scenario_instance(scenario_instance):
    assert isinstance(scenario_instance, Scenario)


def test_scenario_run_folder(scenario_instance, data_folder):
    assert scenario_instance.run_folder == data_folder


def test_scenario_outputfolder(scenario_instance, tmp_folder):
    assert scenario_instance.output_folder == tmp_folder


def test_scenario_mkdirs(scenario_instance, tmp_folder):
    assert scenario_instance.output_folder.exists()
    assert scenario_instance.output_folder == tmp_folder


@pytest.mark.parametrize(
    "input_model, expected_fmap",
    [
        (None, {}),
        ("plexos", read_fmap("r2x/defaults/plexos_mapping.json")),
        ("sienna", read_fmap("r2x/defaults/sienna_mapping.json")),
    ],
)
def test_scenario_fmap(input_model, expected_fmap):
    scenario = Scenario.from_kwargs(name=f"test-{input_model}", input_model=input_model)
    assert scenario.fmap == expected_fmap


def test_scenario_from_kwargs(tmp_folder):
    kwargs = {
        "name": "Test",
        "weather_year": 2015,
        "solve_year": 2055,
        "output_folder": tmp_folder,
        "input_model": "plexos",
        "output_model": "sienna",
        "feature_flags": {"cool-feature": True},
    }
    scenario = Scenario.from_kwargs(**kwargs)
    assert isinstance(scenario, Scenario)
    assert scenario.weather_year == kwargs["weather_year"]
    assert scenario.solve_year == kwargs["solve_year"]
    assert scenario.input_model == kwargs["input_model"]
    assert scenario.output_model == kwargs["output_model"]
    assert scenario.feature_flags == kwargs["feature_flags"]


def test_config_from_cli():
    cli_input = {
        "name": "Test",
        "weather_year": 2015,
        "solve_year": 2055,
        "input_model": "plexos",
        "output_model": "sienna",
        "feature_flags": {"cool-feature": True},
    }
    scenario_mgr = Configuration.from_cli(cli_args=cli_input)

    assert isinstance(scenario_mgr, Configuration)
    assert len(scenario_mgr) == 1


def test_config_override(scenario_instance):
    user_dict = {"fmap": {"xml_file": {"fname": "test_override"}}}
    cli_args = {"output_model": "sienna"}
    scenario = Configuration.override(scenario_instance.__dict__, user_dict=user_dict, cli_args=cli_args)
    assert isinstance(scenario, Scenario)
    assert scenario.output_model == "sienna"
    assert scenario.fmap["xml_file"]["fname"] == "test_override"


def test_configuration_printing(scenario_instance):
    scenario_instance.info()



"""Concrete class to handler parser models.

This module provides the abstract class to create parser objects.
"""

# System packages
from copy import deepcopy
import inspect
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Any, TypeVar
from collections.abc import Callable, Sequence
from pathlib import Path

# Third-party packages
from loguru import logger
import polars as pl
import pandas as pd

# Local packages
from r2x.api import System
from r2x.config import Scenario
from plexosdb import XMLHandler
from .parser_helpers import pl_filter_year, pl_lowercase, pl_rename
from ..utils import check_file_exists


@dataclass
class BaseParser(ABC):
    """Class that defines the shared methods of parsers.

    Note
    ----
    This class is meant to be use for developing new parsers. Do not use it directly.

    Attributes
    ----------
    config: Scenario
        Scenario configuration
    data: dict
        We save each file read in a data dictionary

    Methods
    -------
    get_data(key='load')
        Return the parsed data for load.
    read_file(fpath="load.csv")
        Read load data.
    parse_data
        Read all files from a configuration file map.
    """

    config: Scenario
    data: dict = field(default_factory=dict)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(Files parsed: {len(self.data)})"

    def get_data(self, key: str) -> Any:
        """Return data."""
        if key not in self.data:
            raise KeyError(f"Key `{key}` not found in data dictionary.")
        return self.data[key]

    def read_file(self, fpath: Path | str, filter_funcs: list[Callable] | None = None, **kwargs):
        """Read input model data from the file system.

        Currently supported formats:
            - .csv
            - .h5
            - .xml
        More to come!

        Parameters
        ----------
        fpath: Path, str
            Absolute location of the file in the system
        filter_func: List
            Filter functions to apply

        """
        data = file_handler(fpath, **kwargs)
        if data is None:
            return

        if isinstance(filter_funcs, list):
            for func in filter_funcs:
                data = func(data, **kwargs)
        return data

    def parse_data(
        self,
        *,
        base_folder: str | Path | None,
        fmap: dict,
        filter_func: list[Callable] | None = None,
        **kwargs,
    ) -> None:
        """Parse all the data for the given translation."""
        _fmap = deepcopy(fmap)
        if base_folder is None:
            logger.warning("Missing base folder for {}", self.config.name)
            return None
        logger.trace("Parsing data for {}", self.__class__.__name__)
        for dname, data in _fmap.items():
            if not isinstance(data, dict):
                continue
            if not data.get("fname"):
                continue
            fpath = check_file_exists(fname=data["fname"], run_folder=base_folder)
            if fpath is not None:
                if "fpath" in data:
                    _fpath = data.pop("fpath")
                    # assert fpath == _fpath, f"Multiple files found. {fpath} and {_fpath}"
                    fpath = _fpath
                assert isinstance(fpath, Path) or isinstance(fpath, str)
                fmap[dname]["fpath"] = fpath
                self.data[dname] = self.read_file(fpath=fpath, filter_funcs=filter_func, **{**data, **kwargs})
            logger.debug("Loaded file for {} from {}", dname, fpath)
        return None

    @abstractmethod
    def build_system(self) -> System:
        """Create the infra_sys model."""


class PCMParser(BaseParser):
    """Class defining shared methods for PCM (currently plexos and sienna) parsers."""

    pass


def file_handler(
    fpath: Path | str, optional: bool = False, **kwargs
) -> pl.LazyFrame | Sequence | XMLHandler | None:
    """Return FileHandler based on file extension.

    Raises
    ------
    FileNotFoundError
        If the file is not found.
    NotImplementedError
        If the file format is not yet supported.
    """
    logger.trace("Attempting to read: {}", fpath)
    if not isinstance(fpath, Path):
        fpath = Path(fpath)

    if not fpath.exists() and not optional:
        raise FileNotFoundError(f"Mandatory file {fpath} does not exists.")

    if not fpath.exists() and optional:
        logger.warning("Skipping optional file {}", fpath)
        return None

    match fpath.suffix:
        case ".csv":
            logger.trace("Reading {}", fpath)
            return pl.scan_csv(fpath)
        case ".h5":
            logger.trace("Reading {}", fpath)
            return pl.LazyFrame(pd.read_hdf(fpath).reset_index())  # type: ignore
        case ".xml":
            class_kwargs = {
                key: value for key, value in kwargs.items() if key in inspect.signature(XMLHandler).parameters
            }
            return XMLHandler.parse(fpath=fpath, **class_kwargs)
        case _:
            raise NotImplementedError(f"File {fpath.suffix = } not yet supported.")


ParserClass = TypeVar("ParserClass", bound=BaseParser)


def get_parser_data(
    config: Scenario,
    parser_class: Callable,
    filter_funcs: list[Callable] | None = None,
    **kwargs,
) -> BaseParser:
    """Return parsed system.

    This function will create the ReEDS DataPortal and populate with the most
    common set of data needed


    Paremters
    ---------
    config Scenario configuration class
    parser
        Parser to process
    filter_func
        Functions that will applied to read_data process

    Other Parameters
    ----------------
    kwargs
        year
            For filtering by solve year
        year_column
            To change the column to apply the filter
        column_mapping
            For renaming columns

    See Also
    --------
    BaseParser
    pl_filter_year
    pl_lower_case
    pl_rename
    """
    logger.debug("Creating {} instance.", parser_class.__name__)

    parser = parser_class(config=config, **kwargs)

    # Functions relative to the parser.
    # NOTE: At some point we are going to migrate this out, but this sound like a good standard set
    if filter_funcs is None and config.input_model == "reeds-US":
        logger.trace("Using default filter functions")
        filter_funcs = [pl_lowercase, pl_rename, pl_filter_year]

    # Adding special case for Plexos parser
    if model := getattr(config, "model", False):
        kwargs["model"] = model

    # Parser data
    parser.parse_data(
        fmap=config.fmap,
        base_folder=config.run_folder,
        solve_year=config.solve_year,
        filter_func=filter_funcs,
        **kwargs,
    )

    # Create system
    logger.debug("Starting creation of system: {}", config.name)

    return parser



"""Model related to branches."""

from r2x.models.core import Device
from r2x.models.topology import ACBus, DCBus, Area
from typing import Annotated
from pydantic import Field, NonNegativeFloat, NonPositiveFloat
from r2x.units import ActivePower, Percentage


class Branch(Device):
    """Class representing a connection between components."""

    @classmethod
    def example(cls) -> "Branch":
        return Branch(name="ExampleBranch")


class ACBranch(Branch):
    """Class representing an AC connection between components."""

    # arc: Annotated[Arc, Field(description="The branch's connections.")]
    from_bus: Annotated[ACBus, Field(description="Bus connected upstream from the arc.")]
    to_bus: Annotated[ACBus, Field(description="Bus connected downstream from the arc.")]
    r: Annotated[float, Field(description=("Resistance of the branch"))] = 0
    x: Annotated[float, Field(description=("Reactance of the branch"))] = 0
    b: Annotated[float, Field(description=("Shunt susceptance of the branch"))] = 0
    rating: Annotated[ActivePower, Field(ge=0, description="Thermal rating of the line.")] | None = None


class MonitoredLine(ACBranch):
    """Class representing an AC transmission line."""

    rating_up: Annotated[ActivePower, Field(ge=0, description="Forward rating of the line.")] | None = None
    rating_down: Annotated[ActivePower, Field(le=0, description="Reverse rating of the line.")] | None = None
    losses: Annotated[Percentage, Field(description="Power losses on the line.")] | None = None

    @classmethod
    def example(cls) -> "MonitoredLine":
        return MonitoredLine(
            name="ExampleMonitoredLine",
            from_bus=ACBus.example(),
            to_bus=ACBus.example(),
            losses=Percentage(10, "%"),
            rating_up=ActivePower(100, "MW"),
            rating_down=ActivePower(-100, "MW"),
            rating=ActivePower(100, "MW"),
        )


class Line(ACBranch):
    """Class representing an AC transmission line."""

    @classmethod
    def example(cls) -> "Line":
        return Line(
            name="ExampleLine",
            from_bus=ACBus.example(),
            to_bus=ACBus.example(),
            rating=ActivePower(100, "MW"),
        )


class Transformer2W(ACBranch):
    """Class representing a 2-W transformer."""

    rate: Annotated[NonNegativeFloat, Field(description="Rating of the transformer.")]

    @classmethod
    def example(cls) -> "Transformer2W":
        return Transformer2W(
            name="Example2WTransformer",
            rate=100,
            from_bus=ACBus.example(),
            to_bus=ACBus.example(),
        )


class DCBranch(Branch):
    """Class representing a DC connection between components."""

    from_bus: Annotated[DCBus, Field(description="Bus connected upstream from the arc.")]
    to_bus: Annotated[DCBus, Field(description="Bus connected downstream from the arc.")]


class AreaInterchange(Branch):
    """Collection of branches that make up an interfece or corridor for the transfer of power."""

    max_power_flow: Annotated[ActivePower, Field(ge=0, description="Maximum allowed flow.")]
    min_power_flow: Annotated[ActivePower, Field(le=0, description="Minimum allowed flow.")]
    from_area: Annotated[Area, Field(description="Area containing the bus.")] | None = None
    to_area: Annotated[Area, Field(description="Area containing the bus.")] | None = None

    @classmethod
    def example(cls) -> "AreaInterchange":
        return AreaInterchange(
            name="ExampleAreaInterchange",
            max_power_flow=ActivePower(100, "MW"),
            min_power_flow=ActivePower(-100, "MW"),
            from_area=Area.example(),
            to_area=Area.example(),
        )


class TModelHVDCLine(DCBranch):
    """Class representing a DC transmission line."""

    rating_up: Annotated[NonNegativeFloat, Field(description="Forward rating of the line.")] | None = None
    rating_down: Annotated[NonPositiveFloat, Field(description="Reverse rating of the line.")] | None = None
    losses: Annotated[NonNegativeFloat, Field(description="Power losses on the line.")] = 0
    resistance: Annotated[NonNegativeFloat, Field(description="Resistance of the line in p.u.")] | None = 0
    inductance: Annotated[NonNegativeFloat, Field(description="Inductance of the line in p.u.")] | None = 0
    capacitance: Annotated[NonNegativeFloat, Field(description="Capacitance of the line in p.u.")] | None = 0

    @classmethod
    def example(cls) -> "TModelHVDCLine":
        return TModelHVDCLine(
            name="ExampleDCLine",
            from_bus=DCBus.example(),
            to_bus=DCBus.example(),
            rating_up=100,
            rating_down=80,
        )



"""Create PLEXOS model from translated ReEDS data."""

from typing import Any
import uuid
import string
from collections.abc import Callable

from infrasys.component import Component
from loguru import logger

from r2x.enums import ReserveType
from r2x.exporter.handler import BaseExporter
from plexosdb import PlexosSQLite
from plexosdb.enums import ClassEnum, CollectionEnum
from r2x.models import (
    ACBus,
    Emission,
    InterruptiblePowerLoad,
    Generator,
    GenericBattery,
    HydroDispatch,
    HydroEnergyReservoir,
    HydroPumpedStorage,
    LoadZone,
    MonitoredLine,
    PowerLoad,
    RenewableDispatch,
    RenewableNonDispatch,
    Reserve,
    ThermalStandard,
    Transformer2W,
    TransmissionInterface,
)
from r2x.units import get_magnitude
from r2x.utils import custom_attrgetter, get_enum_from_string, read_json, get_property_magnitude

NESTED_ATTRIBUTES = ["ext", "bus", "services"]
TIME_SERIES_PROPERTIES = ["Min Provision", "Static Risk"]


class PlexosExporter(BaseExporter):
    """Plexos exporter class."""

    def __init__(
        self,
        *args,
        plexos_scenario: str = "default",
        database_manager=None,
        xml_fname: str | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.plexos_scenario = plexos_scenario
        self.property_map = self.config.defaults["plexos_property_map"]
        self.valid_properties = self.config.defaults["valid_properties"]
        self.default_units = self.config.defaults["default_units"]

        self._db_mgr = database_manager or PlexosSQLite(xml_fname=xml_fname)
        self.plexos_scenario_id = self._db_mgr.get_scenario_id(scenario_name=plexos_scenario)

    def run(self, *args, new_database: bool = True, **kwargs) -> "PlexosExporter":
        """Run the exporter."""
        logger.info("Starting {}", self.__class__.__name__)

        self.export_data_files()

        # If starting w/o a reference file we add our custom models and objects
        if new_database:
            self._add_simulation_objects()
            self._add_horizons()
            self._add_models()
            self._add_reports()
        self.add_topology()
        self.add_lines()
        self.add_transformers()
        self.add_interfaces()
        self.add_emissions()
        self.add_reserves()
        self.add_batteries()
        self.add_generators()
        self.add_storage()

        self._db_mgr.to_xml(fpath=f"{self.output_folder}/{self.config.name}.xml")

        return self

    def _get_time_series_properties(self, component):  # noqa: C901
        """Add time series object to certain plexos properties."""
        if not self.system.has_time_series(component):
            return

        ts_metadata = self.system.get_time_series(component)

        config_dict = self.config.__dict__
        config_dict["name"] = self.config.name
        csv_fname = config_dict.get("time_series_fname", "${component_type}_${name}_${weather_year}.csv")
        string_template = string.Template(csv_fname)
        config_dict["component_type"] = f"{component.__class__.__name__}_{ts_metadata.variable_name}"
        csv_fname = string_template.safe_substitute(config_dict)
        csv_fpath = self.ts_directory / csv_fname
        time_series_property: dict[str, Any] = {"Data File": str(csv_fpath)}

        # Property with time series can change. Validate this option based on the component type.
        match component:
            case InterruptiblePowerLoad():
                time_series_property["Fixed Load"] = "0"
            case PowerLoad():
                time_series_property["Load"] = "0"
            case RenewableDispatch():
                time_series_property["Rating"] = "0"
                time_series_property["Load Subtracter"] = "0"
            case RenewableNonDispatch():
                time_series_property["Rating"] = "0"
                time_series_property["Load Subtracter"] = "0"
            case Reserve():
                match component.reserve_type:
                    case ReserveType.SPINNING:
                        time_series_property["Min Provision"] = "0"
                    case ReserveType.FLEXIBILITY:
                        time_series_property["Min Provision"] = "0"
                    case ReserveType.REGULATION:
                        time_series_property["Static Risk"] = "0"
                    case _:
                        raise NotImplementedError(f"Reserve {component.type} not supported")
            case HydroDispatch():
                time_series_property["Max Energy"] = "0"
            case HydroEnergyReservoir():
                time_series_property["Fixed Load"] = "0"
            case ThermalStandard():
                variable_name = self.system.get_time_series(component).variable_name
                if not variable_name:
                    return None
                property_name = self.property_map.get(variable_name, None)
                if property_name:
                    time_series_property[property_name] = "0"
            case _:
                raise NotImplementedError(f"Time Series for {component.label} not supported yet.")
        return time_series_property

    def insert_component_properties(
        self,
        component_type: type["Component"],
        /,
        *,
        parent_class: ClassEnum,
        parent_object_name: str = "System",
        collection: CollectionEnum,
        filter_func: Callable | None = None,
        scenario: str | None = None,
        records: list[dict] | None = None,
        exclude_fields: list[str] = NESTED_ATTRIBUTES,
    ) -> None:
        """Bulk insert properties from selected component type."""
        logger.debug("Adding {} table properties...", component_type.__name__)
        scenario = scenario or self.plexos_scenario
        if not records:
            records = [
                component.model_dump(exclude_none=True, exclude=exclude_fields)
                for component in self.system.get_components(component_type, filter_func=filter_func)
            ]

        if not records:
            logger.warning("No components found for type {}", component_type)
            return

        collection_properties = self._db_mgr.query(
            f"select name, property_id from t_property where collection_id={collection}"
        )
        property_names = [key[0] for key in collection_properties]
        match component_type.__name__:
            case "GenericBattery":
                custom_map = {"base_power": "Max Power", "storage_capacity": "Capacity"}
            case _:
                custom_map = {}
        property_map = self.property_map | custom_map
        valid_component_properties = self.get_valid_records_properties(
            records,
            property_map,
            self.default_units,
            valid_properties=property_names,
        )
        self._db_mgr.add_property_from_records(
            valid_component_properties,
            parent_class=parent_class,
            parent_object_name=parent_object_name,
            collection=collection,
            scenario=scenario,
        )

    def add_component_category(
        self,
        component_type: type["Component"],
        class_enum: ClassEnum,
        category_attribute: str = "category",
        filter_func: Callable | None = None,
    ):
        """Add all categories for a component type."""
        component_categories = {
            custom_attrgetter(component, category_attribute)
            for component in self.system.get_components(component_type, filter_func=filter_func)
        }

        existing_rank = self._db_mgr.query(f"select max(rank) from t_category where class_id = {class_enum}")[
            0
        ][0]
        categories = [
            (class_enum.value, rank, category or "")
            for rank, category in enumerate(sorted(component_categories), start=existing_rank + 1)
        ]

        # Maybe replace `t_category` with right schema.
        with self._db_mgr._conn as conn:
            conn.executemany("INSERT into t_category(class_id, rank, name) values (?,?,?)", categories)
        return

    def bulk_insert_objects(
        self,
        component_type: type["Component"],
        class_enum: ClassEnum,
        collection_enum: CollectionEnum,
        category_attribute: str = "category",
        name_prefix: str = "",
        filter_func: Callable | None = None,
    ):
        """Bulk insert objects to the database."""
        logger.debug("Adding plexos objects for component type {}", component_type.__name__)
        categories_ids = {
            key: value
            for key, value in self._db_mgr.query(
                f"select name, category_id from t_category where class_id = {class_enum}"
            )
        }

        objects = [
            (
                class_enum.value,
                component.name + name_prefix,
                self._get_category_id(component, category_attribute, categories_ids),
                str(uuid.uuid4()),
            )
            for component in self.system.get_components(component_type, filter_func=filter_func)
        ]
        object_names = tuple(d[1] for d in objects)
        if not objects:
            logger.warning("No components found for type {}", component_type)
            return
        with self._db_mgr._conn as conn:
            conn.executemany(
                "INSERT into t_object(class_id, name, category_id, GUID) values (?,?,?,?)", objects
            )

        # Add system membership
        system_object_id: int = self._db_mgr.query("select object_id from t_object where name = 'System'")[0][
            0
        ]
        objects_placeholders = ", ".join("?" * len(object_names))
        membership_query = f"""
            INSERT into t_membership(
              parent_class_id, parent_object_id,
              collection_id, child_class_id, child_object_id
            )
            SELECT
                ? as parent_class_id,
                ? as parent_object_id,
                ? as collection_id,
                ? as child_class_id,
                object_id as child_object_id
            FROM
              t_object
            WHERE
              class_id = ?
              and t_object.name in ({objects_placeholders})
        """
        query_parameters = (
            ClassEnum.System.value,
            system_object_id,
            str(collection_enum),
            class_enum,
            class_enum,
            *object_names,
        )
        self._db_mgr.execute_query(membership_query, query_parameters)

        # Enable all classes that we add into the database.
        # Some are disabled by default.
        self._db_mgr.execute_query(f"UPDATE t_class set is_enabled=1 where class_id={class_enum}")
        return

    def add_topology(self) -> None:
        """Create network topology on Plexos."""
        logger.debug("Adding plexos objects")

        # Adding Regions
        self.add_component_category(ACBus, category_attribute="area.name", class_enum=ClassEnum.Region)
        self.bulk_insert_objects(
            ACBus,
            category_attribute="area.name",
            class_enum=ClassEnum.Region,
            collection_enum=CollectionEnum.SystemRegions,
        )
        self.insert_component_properties(
            ACBus, parent_class=ClassEnum.System, collection=CollectionEnum.SystemRegions
        )

        # Adding Zones
        # self.add_component_category(LoadZone, class_enum=ClassEnum.Zone)
        self.bulk_insert_objects(
            LoadZone, class_enum=ClassEnum.Zone, collection_enum=CollectionEnum.SystemZones
        )
        self.insert_component_properties(
            LoadZone, parent_class=ClassEnum.System, collection=CollectionEnum.SystemZones
        )

        # Adding nodes
        # NOTE: For nodes, we do not add category.
        # self.add_component_category(ACBus, class_enum=ClassEnum.Node)
        self.bulk_insert_objects(ACBus, class_enum=ClassEnum.Node, collection_enum=CollectionEnum.SystemNodes)
        self.insert_component_properties(
            ACBus, parent_class=ClassEnum.System, collection=CollectionEnum.SystemNodes
        )

        # Add node memberships to zone and regions.
        # On our default Plexos translation, both Zones and Regions are child of the Node class.
        for bus in self.system.get_components(ACBus):
            self._db_mgr.add_membership(
                bus.name,
                bus.name,  # Zone has the same name
                parent_class=ClassEnum.Node,
                child_class=ClassEnum.Region,
                collection=CollectionEnum.NodesRegion,
            )
            self._db_mgr.add_membership(
                bus.name,
                bus.load_zone.name,
                parent_class=ClassEnum.Node,
                child_class=ClassEnum.Zone,
                collection=CollectionEnum.NodesZone,
            )

        # Adding load time series
        logger.debug("Adding load time series properties")
        for component in self.system.get_components(
            PowerLoad, filter_func=lambda component: self.system.has_time_series(component)
        ):
            time_series_properties = self._get_time_series_properties(component)
            if time_series_properties:
                text = time_series_properties.pop("Data File")
                for property_name, property_value in time_series_properties.items():
                    self._db_mgr.add_property(
                        component.bus.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Region,
                        collection=CollectionEnum.SystemRegions,
                        scenario=self.plexos_scenario,
                        text={"Data File": text},
                    )
        return

    def add_lines(self) -> None:
        """Add transmission lines that connect topology elements."""
        # Adding Lines
        # NOTE: The default line on Plexos is a `MonitoredLine` without category. If we need to add a category
        # in the future, we will uncomment the line below with the pertinent category name.
        # self.add_component_category(MonitoredLine, class_enum=ClassEnum.Line)
        self.bulk_insert_objects(
            MonitoredLine, class_enum=ClassEnum.Line, collection_enum=CollectionEnum.SystemLines
        )
        self.insert_component_properties(
            MonitoredLine, parent_class=ClassEnum.System, collection=CollectionEnum.SystemLines
        )

        for line in self.system.get_components(MonitoredLine):
            properties = self.get_valid_component_properties(
                line.ext,
                property_map=self.property_map,
                unit_map=self.default_units,
                collection=CollectionEnum.SystemLines,
            )
            for property_name, property_value in properties.items():
                self._db_mgr.add_property(
                    line.name,
                    property_name,
                    property_value,
                    object_class=ClassEnum.Line,
                    collection=CollectionEnum.SystemLines,
                    scenario=self.plexos_scenario,
                )
            self._db_mgr.add_membership(
                line.name,
                line.from_bus.name,
                parent_class=ClassEnum.Line,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.LineNodeFrom,
            )
            self._db_mgr.add_membership(
                line.name,
                line.to_bus.name,
                parent_class=ClassEnum.Line,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.LineNodeTo,
            )
        return

    def add_transformers(self) -> None:
        """Add Transformer objects to the database."""
        self.add_component_category(Transformer2W, class_enum=ClassEnum.Transformer)
        self.bulk_insert_objects(
            Transformer2W, class_enum=ClassEnum.Transformer, collection_enum=CollectionEnum.SystemTransformers
        )
        self.insert_component_properties(
            Transformer2W, parent_class=ClassEnum.System, collection=CollectionEnum.SystemTransformers
        )
        for transformer in self.system.get_components(Transformer2W):
            self._db_mgr.add_membership(
                transformer.name,
                transformer.from_bus.name,
                parent_class=ClassEnum.Transformer,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.TransformerNodeFrom,
            )
            self._db_mgr.add_membership(
                transformer.name,
                transformer.to_bus.name,
                parent_class=ClassEnum.Transformer,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.TransformerNodeTo,
            )
        return

    def add_interfaces(self) -> None:
        """Add transmission interfaces."""
        self.bulk_insert_objects(
            TransmissionInterface,
            class_enum=ClassEnum.Interface,
            collection_enum=CollectionEnum.SystemInterfaces,
        )
        self.insert_component_properties(
            TransmissionInterface, parent_class=ClassEnum.System, collection=CollectionEnum.SystemInterfaces
        )

    def add_emissions(self) -> None:
        """Add emission objects to the database."""
        # Getting all unique emission types (e.g., CO2, NOX) from the emissions objects.
        # NOTE: On Plexos, we need to add each emission type individually to the Emission class
        self._db_mgr.execute_query(
            f"UPDATE t_class set is_enabled=1 where class_id={ClassEnum.Emission.value}"
        )
        emission_types = set(map(lambda x: x.emission_type, list(self.system.get_components(Emission))))
        for emission_type in emission_types:
            self._db_mgr.add_object(
                emission_type,
                ClassEnum.Emission,
                CollectionEnum.SystemEmissions,
            )

        return

    def add_reserves(self) -> None:
        """Add system reserves to the database.

        This method only appends the information for the type of reserves. Each
        generator that contributes to the reserve will add the membership for
        the type of reserve as a child membership.
        """
        # Adding reserves
        # self.add_component_category(Reserve, class_enum=ClassEnum.Reserve)
        self.bulk_insert_objects(
            Reserve, class_enum=ClassEnum.Reserve, collection_enum=CollectionEnum.SystemReserves
        )

        self.insert_component_properties(
            Reserve,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.SystemReserves,
            exclude_fields=[*NESTED_ATTRIBUTES, "max_requirement"],
        )
        for reserve in self.system.get_components(Reserve):
            properties = {}
            properties["Type"] = reserve.ext["Type"]
            properties["Is Enabled"] = "-1" if reserve.available else "0"
            properties["Mutually Exclusive"] = True

            for property, value in properties.items():
                self._db_mgr.add_property(
                    reserve.name,
                    property,
                    value,
                    object_class=ClassEnum.Reserve,
                    collection=CollectionEnum.SystemReserves,
                    scenario=self.plexos_scenario,
                )
            time_series_properties = self._get_time_series_properties(reserve)
            if time_series_properties:
                text = time_series_properties.pop("Data File")
                for property_name, property_value in time_series_properties.items():
                    self._db_mgr.add_property(
                        reserve.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Reserve,
                        collection=CollectionEnum.SystemReserves,
                        scenario=self.plexos_scenario,
                        text={"Data File": text},
                    )

            # Add ReserveRegions properties. Currently, we only add the load_risk
            component_dict = reserve.model_dump(
                exclude_none=True, exclude=[*NESTED_ATTRIBUTES, "max_requirement"]
            )

            regions = self.system.get_components(
                ACBus, filter_func=lambda x: x.load_zone.name == reserve.region.name
            )

            for region in regions:
                self._db_mgr.add_membership(
                    reserve.name,
                    region.name,  # Zone has the same name
                    parent_class=ClassEnum.Reserve,
                    child_class=ClassEnum.Region,
                    collection=CollectionEnum.ReserveRegions,
                )
                properties = self.get_valid_component_properties(
                    component_dict,
                    property_map=self.property_map,
                    unit_map=self.default_units,
                    collection=CollectionEnum.ReserveRegions,
                )
                if properties:
                    for property_name, property_value in properties.items():
                        self._db_mgr.add_property(
                            region.name,
                            property_name,
                            property_value,
                            object_class=ClassEnum.Region,
                            parent_object_name=reserve.name,
                            parent_class=ClassEnum.Reserve,
                            collection=CollectionEnum.ReserveRegions,
                            scenario=self.plexos_scenario,
                        )
        return

    def add_generators(self):
        """Add generator objects to the database."""

        # Add generator objects excluding batteries
        def exclude_battery(component):
            return not isinstance(component, GenericBattery)

        self.add_component_category(Generator, class_enum=ClassEnum.Generator, filter_func=exclude_battery)
        self.bulk_insert_objects(
            Generator,
            class_enum=ClassEnum.Generator,
            collection_enum=CollectionEnum.SystemGenerators,
            filter_func=exclude_battery,
        )
        self.insert_component_properties(
            Generator,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.SystemGenerators,
            filter_func=exclude_battery,
        )

        # Add generator memberships
        logger.debug("Adding generator memberships")
        for generator in self.system.get_components(Generator, filter_func=exclude_battery):
            self._db_mgr.add_membership(
                generator.name,
                generator.bus.name,
                parent_class=ClassEnum.Generator,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.GeneratorNodes,
            )
            properties = self._get_time_series_properties(generator)
            if properties:
                text = properties.pop("Data File")
                for property_name, property_value in properties.items():
                    self._db_mgr.add_property(
                        generator.name,
                        property_name,
                        property_value,
                        object_class=ClassEnum.Generator,
                        collection=CollectionEnum.SystemGenerators,
                        scenario=self.plexos_scenario,
                        text={"Data File": text},
                    )
            if generator.services:
                for service in generator.services:
                    match service:
                        case Reserve():
                            self._db_mgr.add_membership(
                                service.name,
                                generator.name,
                                parent_class=ClassEnum.Reserve,
                                child_class=ClassEnum.Generator,
                                collection=CollectionEnum.ReserveGenerators,
                            )
                        case _:
                            raise NotImplementedError(f"{service} not yet implemented for generator.")

        # NOTE: This needs to be optimized. It is currently slow.
        logger.debug("Adding generator emisssions memberships")
        for emission in self.system.get_components(Emission):
            self._db_mgr.add_membership(
                emission.emission_type,
                emission.generator_name,
                parent_class=ClassEnum.Emission,
                child_class=ClassEnum.Generator,
                collection=CollectionEnum.EmissionGenerators,
            )
            self._db_mgr.add_property(
                emission.generator_name,
                self.property_map["rate"],
                get_magnitude(emission.rate),
                object_class=ClassEnum.Generator,
                parent_class=ClassEnum.Emission,
                parent_object_name=emission.emission_type,
                collection=CollectionEnum.EmissionGenerators,
                scenario=self.plexos_scenario,
            )

    def add_batteries(self):
        """Add battery objects to the database."""
        # Add battery objects
        self.add_component_category(GenericBattery, class_enum=ClassEnum.Battery)
        self.bulk_insert_objects(
            GenericBattery,
            class_enum=ClassEnum.Battery,
            collection_enum=CollectionEnum.SystemBattery,
        )
        self.insert_component_properties(
            GenericBattery, parent_class=ClassEnum.System, collection=CollectionEnum.SystemBattery
        )
        # Add battery memberships
        for battery in self.system.get_components(GenericBattery):
            self._db_mgr.add_membership(
                battery.name,
                battery.bus.name,
                parent_class=ClassEnum.Battery,
                child_class=ClassEnum.Node,
                collection=CollectionEnum.BatteryNodes,
            )
            if battery.services:
                for service in battery.services:
                    match service:
                        case Reserve():
                            self._db_mgr.add_membership(
                                service.name,
                                battery.name,
                                parent_class=ClassEnum.Reserve,
                                child_class=ClassEnum.Battery,
                                collection=CollectionEnum.ReserveBatteries,
                            )
                        case _:
                            raise NotImplementedError(f"{service} not yet implemented for generator.")

    def add_storage(self):
        """Add storage objects to the database."""
        # Add pump storage objects
        self.add_component_category(
            HydroPumpedStorage, category_attribute="head", class_enum=ClassEnum.Storage
        )
        self.add_component_category(
            HydroPumpedStorage, category_attribute="tail", class_enum=ClassEnum.Storage
        )
        self.bulk_insert_objects(
            HydroPumpedStorage,
            class_enum=ClassEnum.Storage,
            category_attribute="head",
            collection_enum=CollectionEnum.SystemStorage,
            name_prefix="_head",
        )
        self.bulk_insert_objects(
            HydroPumpedStorage,
            class_enum=ClassEnum.Storage,
            category_attribute="tail",
            collection_enum=CollectionEnum.SystemStorage,
            name_prefix="_tail",
        )

        head_storage = [
            component.model_dump(exclude_none=True, exclude=NESTED_ATTRIBUTES)
            for component in self.system.get_components(HydroPumpedStorage)
        ]
        for component in head_storage:
            component["name"] += "_head"
        self.insert_component_properties(
            HydroPumpedStorage,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.SystemStorage,
            records=head_storage,
        )
        tail_storage = [
            component.model_dump(exclude_none=True, exclude=NESTED_ATTRIBUTES)
            for component in self.system.get_components(HydroPumpedStorage)
        ]
        for component in tail_storage:
            component["name"] += "_tail"
        self.insert_component_properties(
            HydroPumpedStorage,
            parent_class=ClassEnum.System,
            collection=CollectionEnum.SystemStorage,
            records=tail_storage,
        )

        for phs in self.system.get_components(HydroPumpedStorage):
            head_name = f"{phs.name}_head"
            tail_name = f"{phs.name}_tail"
            self._db_mgr.add_membership(
                phs.name,
                head_name,
                parent_class=ClassEnum.Generator,
                child_class=ClassEnum.Storage,
                collection=CollectionEnum.GeneratorHeadStorage,
            )
            self._db_mgr.add_membership(
                phs.name,
                tail_name,
                parent_class=ClassEnum.Generator,
                child_class=ClassEnum.Storage,
                collection=CollectionEnum.GeneratorTailStorage,
            )
        return

    def _add_simulation_objects(self):
        for simulation_object in self.config.defaults["simulation_objects"]:
            collection_enum = get_enum_from_string(simulation_object["name"], CollectionEnum, prefix="System")
            class_enum: ClassEnum = get_enum_from_string(simulation_object["name"], ClassEnum)
            for objects in simulation_object["attributes"]:
                self._db_mgr.add_object(
                    objects["name"],
                    class_enum,
                    collection_enum,
                    category_name=objects["category"] or "-",
                )

            # Add properties
            property_list = self.config.defaults[simulation_object["name"]]
            for attributes in property_list["attributes"]:
                self._db_mgr.add_attribute(
                    object_name=attributes["name"],
                    object_class=get_enum_from_string(attributes["class"], ClassEnum),
                    attribute_class=class_enum,
                    attribute_name=attributes["attribute"],
                    attribute_value=attributes["value"],
                )

    def _add_horizons(self):
        logger.info("Adding model horizon")
        static_horizon_type = self.config.defaults["static_horizon_type"]
        static_horizons = self.config.defaults[static_horizon_type]
        for horizon, values in static_horizons.items():
            self._db_mgr.add_object(
                horizon,
                ClassEnum.Horizon,
                CollectionEnum.SystemHorizon,
            )
            for attribute, attribute_value in values["attributes"].items():
                self._db_mgr.add_attribute(
                    object_name=horizon,
                    object_class=ClassEnum.Horizon,
                    attribute_class=ClassEnum.Horizon,
                    attribute_name=attribute,
                    attribute_value=attribute_value,
                )

    def _add_models(self):
        static_model_type = self.config.defaults["static_model_type"]
        static_models = self.config.defaults[static_model_type]
        for model, values in static_models.items():
            self._db_mgr.add_object(
                model,
                ClassEnum.Model,
                CollectionEnum.SystemModel,
                category_name=values["category"],
            )
            for attribute, attribute_value in values["attributes"].items():
                self._db_mgr.add_attribute(
                    object_name=model,
                    object_class=ClassEnum.Model,
                    attribute_class=ClassEnum.Model,
                    attribute_name=attribute,
                    attribute_value=attribute_value,
                )
            for child_class_name, child_object_name in values["memberships"].items():
                collection_enum = get_enum_from_string(child_class_name, CollectionEnum, prefix="Model")
                child_class = get_enum_from_string(child_class_name, ClassEnum)
                self._db_mgr.add_membership(
                    model,
                    child_object_name,
                    parent_class=ClassEnum.Model,
                    child_class=child_class,
                    collection=collection_enum,
                )
            self._db_mgr.add_membership(
                model,
                self.plexos_scenario,
                parent_class=ClassEnum.Model,
                child_class=ClassEnum.Scenario,
                collection=CollectionEnum.ModelScenario,
            )
        return

    def _add_reports(self):
        fpath = self.config.defaults["plexos_reports"]
        report_objects = read_json(fpath)

        for report_object in report_objects:
            self._db_mgr.add_report(**report_object)
        return

    def _get_category_id(self, component, category_attribute, categories_ids, default_category="-"):
        category_to_get = custom_attrgetter(component, category_attribute)

        if category_to_get not in categories_ids:
            return categories_ids[default_category]
        return categories_ids[category_to_get]

    def get_valid_component_properties(
        self,
        component_dict: dict,
        property_map: dict[str, str],
        unit_map: dict[str, str],
        collection: CollectionEnum,
    ):
        """Validate single component properties."""
        valid_component_properties = {}
        component_dict_mapped = {property_map.get(key, key): value for key, value in component_dict.items()}
        collection_properties = self._db_mgr.query(
            f"select name, property_id from t_property where collection_id={collection}"
        )
        valid_properties = [key[0] for key in collection_properties]
        for property_name, property_value in component_dict_mapped.items():
            if property_name in valid_properties:
                property_value = get_property_magnitude(property_value, to_unit=unit_map.get(property_name))
                valid_component_properties[property_name] = property_value
        return valid_component_properties



from r2x.enums import PrimeMoversType
from r2x.models import Generator, ACBus, Emission, HydroPumpedStorage, ThermalStandard
from r2x.units import EmissionRate, ureg


def test_generator_model():
    generator = ThermalStandard.example()
    assert isinstance(generator, ThermalStandard)
    assert isinstance(generator.active_power.magnitude, float)


def test_emission_objects():
    emission = Emission.example()

    assert isinstance(emission, Emission)
    assert isinstance(emission.rate, EmissionRate)
    assert isinstance(emission.emission_type, str)
    assert emission.emission_type == "CO2"


def test_bus_model():
    bus = ACBus.example()
    assert isinstance(bus, ACBus)
    assert isinstance(bus.number, int)


def test_generator_objects():
    bus = ACBus.example()
    generator = Generator(name="GEN01", active_power=100 * ureg.MW, bus=bus)
    assert isinstance(generator.bus, ACBus)


def test_pumped_hydro_generator():
    pumped_storage = HydroPumpedStorage.example()
    assert isinstance(pumped_storage, HydroPumpedStorage)
    assert isinstance(pumped_storage.bus, ACBus)
    assert isinstance(pumped_storage.prime_mover_type, PrimeMoversType)
    assert pumped_storage.prime_mover_type == PrimeMoversType.PS
